from ._response import FlyMyAIResponse
from .models import (
    FlyMyAI401Response,
    FlyMyAI422Response,
    Base4xxResponse,
    FlyMyAI400Response,
)


class BaseFlyMyAIException(Exception):
    msg: str
    requires_retry: bool
    _response: FlyMyAIResponse

    def __init__(self, msg, requires_retry=False, response=None):
        self.msg = msg
        self.requires_retry = requires_retry
        self._response = response

    @property
    def response(self):
        return self._response

    @classmethod
    def from_5xx(cls, response: FlyMyAIResponse):
        msg = f"""
                INTERNAL SERVER ERROR ({response.status_code}):
                REQUEST URL: {response.url};
            """
        internal_error_mapping = {
            500: lambda: cls(msg, False, response=response),
            502: lambda: cls(msg, True, response=response),
            503: lambda: cls(msg, False, response=response),
            504: lambda: cls(msg, True, response=response),
            524: lambda: cls(msg, True, response=response),
        }
        return internal_error_mapping.get(
            response.status_code, lambda: cls(msg, False)
        )()

    @classmethod
    def from_4xx(cls, response: FlyMyAIResponse):
        response_validation_templates = {
            400: FlyMyAI400Response,
            401: FlyMyAI401Response,
            422: FlyMyAI422Response,
        }
        response_4xx = response_validation_templates.get(
            response.status_code, Base4xxResponse
        ).from_response(response)
        return cls(
            msg=response_4xx.to_msg(),
            requires_retry=response_4xx.requires_retry,
            response=response,
        )

    @classmethod
    def from_response(cls, response: FlyMyAIResponse):
        if 400 <= response.status_code < 500:
            return cls.from_4xx(response)
        if response.status_code >= 500:
            return cls.from_5xx(response)

    def __str__(self):
        return self.msg


class FlyMyAIPredictException(BaseFlyMyAIException):
    ...


class FlyMyAIOpenAPIException(BaseFlyMyAIException):
    ...


class FlyMyAIExceptionGroup(Exception):
    def __init__(self, errors: list[BaseFlyMyAIException], **kwargs):
        self.errors = errors
        exceptions_message = ";".join([str(err) for err in errors])
        self.message = f"FlyMyAI exception history: {exceptions_message}"
        super().__init__(self.message)

    def __str__(self):
        return self.message



import httpx


class FlyMyAIResponse(httpx.Response):
    @classmethod
    def from_httpx(cls, response: httpx.Response):
        return cls(
            status_code=response.status_code,
            content=response.content,
            request=response.request,
            headers=response.headers,
        )



from typing import Any


class BaseField(object):
    def __init__(self, value, name=None):
        self.name = name
        self.value = value

    def validate(self, value: Any = None):
        raise NotImplementedError

    def serialize(self, value: Any = None):
        self.validate(value)
        return value



from typing import Callable, Awaitable, Type, Any, Coroutine

import httpx


def retryable_callback(
    cb: Callable,
    retries: int,
    append_on_exception_cls: Type[Exception],
    exception_group_cls: Type[Exception],
):
    """
    Decorator to retry a function
    """
    retries_history = []
    for _ in range(retries):
        try:
            res = cb()
            return retries_history, res
        except append_on_exception_cls as e:
            retries_history.append(e)
            if e.requires_retry:
                continue
            else:
                raise exception_group_cls(retries_history)
    else:
        exception_gr = exception_group_cls(retries_history)
        raise exception_gr


async def aretryable_callback(
    cb: Callable[..., Awaitable[httpx.Response]],
    retries,
    append_on_exception_cls: Type[Exception],
    exception_group_cls: Type[Exception],
):
    """
    Decorator to retry a function
    """
    retries_history = []
    for _ in range(retries):
        try:
            res = await cb()
            return retries_history, res
        except append_on_exception_cls as e:
            retries_history.append(e)
            if e.requires_retry:
                continue
            else:
                raise exception_group_cls(retries_history)
    else:
        exception_gr = exception_group_cls(retries_history)
        raise exception_gr



import dataclasses

import httpx


class ClientInfo:

    """
    Base class for all ClientInfo objects
    """

    @property
    def authorization_headers(self):
        raise NotImplemented()

    @property
    def _project_path(self):
        raise NotImplemented

    @property
    def prediction_path(self):
        raise NotImplemented

    @property
    def openapi_schema_path(self):
        raise NotImplemented


@dataclasses.dataclass
class APIKeyClientInfo(ClientInfo):

    """
    Encapsulates information about a project.
    Uses X-API-KEY header to perform an auth
    """

    apikey: str
    username: str
    project_name: str

    @property
    def authorization_headers(self):
        return {"X-API-KEY": self.apikey}

    @property
    def _project_path(self):
        return httpx.URL(f"/api/v1/{self.username}/{self.project_name}/")

    @property
    def prediction_path(self):
        return self._project_path.join(httpx.URL("predict"))

    @property
    def openapi_schema_path(self):
        return self._project_path.join(httpx.URL("openapi.json"))


class ClientInfoFactory:
    _raw_auth: dict

    def __init__(self, raw_auth: dict):
        self._raw_auth = raw_auth

    def _build_auth(self) -> ClientInfo:
        """
        Build authorization
        """
        if "apikey" in self._raw_auth:
            return APIKeyClientInfo(**self._raw_auth)
        else:
            raise NotImplemented("This type of authorization is not implemented yet!")

    def build_auth(self):
        return self._build_auth()
